unit uMainForm;

{******************************************************************************
Initial Author   :    Scott Price
Initially Created:    18-Oct-2002
Language         :    Object Pascal
Compiler Version :    Delphi 3.02, 4.03, 5.01, 6.02 (UK)
Copyright        :     Information Technology Sector Limited, 1996-2002

Source Code Constraints:
========================
   Distribution, disclosure or modifications of this source code, outside of
   any licenced agreements with Information Technology Sector Limited, or
   removal of this notice constitutes a breach of any or all license agreements.

Operating Systems Tested On:
============================
   Microsoft Windows(R) 98, First Edition
   Microsoft Windows(R) NT4, Service Pack 6a

Design Notes:
=============

VERSION  AUTHOR                DATE/TIME
===============================================================================
$history$
$nokeywords$
===============================================================================
NOTES:
 - This uses an example class/unit that shows off some of the options for
   using the JclDebug unit.
 - This is by no means a comprehensive demonstration - just read the units
   source code, and you will see what I mean.
 - There is Help for the JCL in general, but it is very limited in some cases
   as we all love to write code, but we hate documentation.  Well, they are
   always keen for any information from people happy to contribute to either,
   but the documentation side is probably more needed.  ;) ;)
===============================================================================
TODO:
******************************************************************************}


interface



uses
  { Own Units }
  uJclExceptInfo,  // This is the unit containing the encapsulation of the JclDebug use.

  { Borland Standard Units }
{$IFDEF VER140}
  Variants,
{$ENDIF}
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls, ComCtrls;


type
  { TfmMain }

  TfmMain = class(TForm)
    pnlButtons: TPanel;
    btnAddToList: TButton;
    moInformation: TMemo;
    stbInformation: TStatusBar;
    btnDoStartMe: TButton;
    btnRememberRemember: TButton;
    procedure btnAddToListClick(Sender: TObject);
    procedure btnDoStartMeClick(Sender: TObject);
    procedure btnRememberRememberClick(Sender: TObject);
  private
    { Private declarations }
    FList: TList;
    { This Object is based on the Class that Encapsulates the use of the JclDebug facilities }
    FExceptionStuff: uJclExceptInfo.TStandardExceptionInfo;
  protected
    { Protected declarations }
    procedure DisplayExceptionInfo(Sender: TObject; const Info: TStrings); virtual;
    procedure WriteOutExceptionInfo(Sender: TObject; const Info: TStrings); virtual;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  end;


var
  fmMain: TfmMain;


implementation

{$R *.dfm}


{ This is the visual example that comes with the JCL, but actually if you look
  at the code you will see why this version won't work with any non-visual
  applications like services.  It requires the Application object's OnException
  event handler to instanitae it.  However, feel free to comment it out and try
  compiling with it also.  You should basically get this same exception
  information in a pre-constructed visual form without having to create or
  Hook my exception object before you generate the exception.  Both approaches
  have their merits and there uses, just determine which suits you most  :) }
(*uses
  ExceptDlg;*)


{ TfmMain }

constructor TfmMain.Create(AOwner: TComponent);
const
  sInfo = ' - This is by no means a comprehensive demonstration - just read the units source code, and you will see what I mean.'#13#10#13#10 +
          ' - There is Help for the JCL in general, but it is very limited in some cases as we all love to write code, but we hate documentation.  Well, they are ' +
          'always keen for any information from people happy to contribute to either, but the documentation side is probably more needed.  ;) ;)'#13#10#13#10;
begin
  inherited Create(AOwner);

  { Here, as you can see from the commented out line, I have created it in such
    a way that you can provide an exception handler in the CreateHooked
    constructor that automatically sets up the handler.  Your choice as to your
    preferred use }
  { FExceptionStuff := TStandardExceptionInfo.CreateHooked(Self.DisplayExceptionInfo); }
  FExceptionStuff := TStandardExceptionInfo.Create;
  FExceptionStuff.OnException := Self.DisplayExceptionInfo;

  { This is Deliberate.  This is the cause of our Exception, a Nil Pointer }
  FList := Nil;

  moInformation.Lines.Add(sInfo);
end;

destructor TfmMain.Destroy;
begin
  TRY
    { Finally, release stuff.  Again, in this instance a deliberate error.
      Whilst you won't see it this time, this does still trap the exception
      and log the information.  The Correct way to destroy this object would
      actually be to call Free.  Free checks to see if the caller is Nil, and
      if it is does not call Destroy.  If the caller is not Nil, destroy will
      be called.

      HOWEVER:  Free can still fail if you have destroyed an object and not
                set the pointer back to Nil!  So a good practise is always to
                call Free and immediately set the object pointer to Nil, or to
                call the FreeAndNil routine if in your version of Delphi, which
                does both in one line of code (well our code) }

    { Ok, so because of this deliberate exception being planned, how do we get
     this useful information still?  Well, we set-up the Exception handling
     object to WriteOut this information to a file instead, as we otherwise
     wouldn't see this exception information }
    FExceptionStuff.OnException := WriteOutExceptionInfo;
    TRY
      FList.Destroy;
      (* FList.Free; // Try with this line instead of the above - you won't get an error *)
    EXCEPT
      { If we didnt' have this TRY..EXCEPT block here, the exception would
        just propogate to the TRY..FINALLY block that would free the resources
        before traping the exception information at the highest level, the
        un-handled exception handler - just something to note.

        This TRY..FINALLY is my prefered way of actually controlling resource
        cleaning in a destructor normally, but in this case would actually
        prevent you getting this useful information without this additional
        TRY..EXCEPT block - but will prevent a memory leak of the object not
        being released }
    END;
  FINALLY
    if FExceptionStuff <> Nil then
      FreeAndNil(FExceptionStuff);

    inherited Destroy;
  END;
end;

procedure TfmMain.DisplayExceptionInfo(Sender: TObject; const Info: TStrings);
const
  scExceptionHappened = 'ARH!!:  Nasty Exception Thingies!  Better Fix ME!!';
begin
  { Do stuff with this lovely information }
  stbInformation.SimpleText := scExceptionHappened;

  { In this case we are just passing this information to the Memo for display
    on screen to the user (well, to us) }
  moInformation.Lines.AddStrings(Info);
end;

procedure TfmMain.WriteOutExceptionInfo(Sender: TObject;
  const Info: TStrings);
var
  sLogFileName: String;
  fsOutput: TFileStream;
begin
  { Ok, here we want to write out the exception information to a file instead
    of trying to display it on screen, so that we catch the closing exceptions }
  sLogFileName := ChangeFileExt(Application.ExeName, '.log');

  TRY
    fsOutput := TFileStream.Create(sLogFileName, fmCreate);
    TRY
      fsOutput.Position := 0;

      Info.SaveToStream(fsOutput);
    FINALLY
      { Correct way to release an object.  Either approach is fine. }
      (* fsOutput.Free;
      fsOutput := Nil; *)
      FreeAndNil(fsOutput);
    END;
  EXCEPT
    { Err... Well, just exit }
    Exit;
  END;
end;

procedure TfmMain.btnAddToListClick(Sender: TObject);
begin
  { This is the line that generates the exception - or is it? }
  FList.Add(Self);

  { If you look at the information in the output you will see that yes this
    line is in the call stack, but it isn't the line where the exception had
    occured.  What is happening is that you are trying to call a Classes Method,
    but the object reference is Nil.  So when it get into the TList.Add method
    and tries to get information from a Field the object should have information
    in it creates the exception.  So actually the exception happens inside
    CLASSES.pas in the first few lines:

function TList.Add(Item: Pointer): Integer;
begin
  Result := FCount;  <-- Here is where the exception is acutally generated :)

    So you can see that this (whilst it occurs because the FList variable is Nil
    isn't actually genereate at the FList.Add line of code, but instead inside
    the TList.Add method trying to opperate on FList.  I hope that wasn't too
    confusing :) }
end;

procedure TfmMain.btnDoStartMeClick(Sender: TObject);
begin
  { Well well well, don't forget to start hooking for the Exception Info first!
    Otherwise, strangely enough, you will not see any information  ;) }
  if NOT FExceptionStuff.Hooked then
    FExceptionStuff.HookExceptions;
end;

procedure TfmMain.btnRememberRememberClick(Sender: TObject);
const
  scTheMessage = 'Remember Remember:'#13#13 +
                 '1)  To Compile/Build with DEBUG DCU''s.'#13 +
                 '2)  a) Compile in the JclDebug info or'#13 +
                 '     b) Compile in the TurboDebug symbols or'#13 +
                 '     c) Deploy the Map File generated at compile time or'#13 +
                 '     d) Deploy the JclDebug Info Converted Map File';
begin
  { Dont forget the possible ways that this information can be obtained
    from the JclDebug units interogation.  These variations allow you to be
    more flexible about how you compile your application, or allow the users
    to run the debug options to provide you with more information }
  ShowMessage(scTheMessage);
end;

end.
